import time
import typing

from reviser import definitions
from ..deploying import publisher
from ..deploying import uploader


def deploy_target(
    target: "definitions.Target",
    description: str = None,
    published_layers: typing.List["definitions.PublishedLayer"] = None,
    dry_run: bool = False,
) -> typing.List["definitions.PublishedLayer"]:
    """
    Deploy the bundled ZIP file to S3 for access by Lambda functions. The
    file is uploaded to the specified storage bucket with a timestamped key
    and then published to the lambda function.

    :param target:
        Configuration target to deploy.
    """
    s3_keys = uploader.upload(target, dry_run=dry_run)

    # Sleep for a few seconds to make sure the S3 upload has finished
    # processing to prevent race conditions during deployment
    if not dry_run:
        print("[SYNC]: Synchronization block between S3 and lambda")
        time.sleep(2)

    if target.kind == definitions.TargetType.LAYER:
        return publisher.publish_layer(
            target=target,
            s3_keys=s3_keys,
            description=description,
            dry_run=dry_run,
        )

    publisher.publish_function(
        target=target,
        s3_keys=s3_keys,
        published_layers=published_layers or [],
        description=description,
        dry_run=dry_run,
    )
    return []


def deploy(
    context: "definitions.Context",
    selection: "definitions.Selection",
    description: str = None,
    dry_run: bool = False,
) -> typing.List["definitions.Target"]:
    """
    Executes the deploy operation on the context's configuration filtered
    by the specified selection.
    """
    selected = context.get_selected_targets(selection)

    # Publish layers first so that the function publishes can update to
    # use any newly published layers.
    published_layers = []
    for target in selected.layer_targets:
        published_layers += deploy_target(
            target=target,
            description=description,
            published_layers=None,
            dry_run=dry_run,
        )

    for target in selected.function_targets:
        deploy_target(
            target=target,
            description=description,
            published_layers=published_layers,
            dry_run=dry_run,
        )

    return selected.targets



"""
Allows for selecting subsets of the targets within the loaded configuration.
The subsets are fuzzy-matched unless the --exact flag is used.
"""
import argparse
import dataclasses
import typing

from reviser import definitions
from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Shell auto-completes for this command."""
    return ["*", "function", "layer", "--exact"]


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for the select command."""
    parser.add_argument(
        "kind",
        choices=[
            definitions.TargetType.FUNCTION.value,
            definitions.TargetType.LAYER.value,
            "*",
        ],
        help="""
            Specify the type of objects to select or use * to
            select both types.
            """,
    )
    parser.add_argument(
        "name",
        nargs="?",
        help="""
        Specifies the value to match against the function
        and layer target names available from the configuration.
        This can include shell-style wildcards and will also
        match against partial strings. If the --exact flag is
        specified, this value must exactly match one of the targets
        instead of the default fuzzy matching behavior.
        """,
    )
    parser.add_argument(
        "--exact",
        action="store_true",
        help="Forces the match to be exact instead of fuzzy.",
    )


def _update_exact_selection(
    kind: str,
    names: typing.List[str],
    selection: "definitions.Selection",
) -> "definitions.Selection":
    """Create a new selection object updated with the specified exact match."""
    layer_names = []
    if kind in ("*", "layer"):
        layer_names = list(names)

    function_names = []
    if kind in ("*", "function"):
        function_names = list(names)

    return dataclasses.replace(
        selection,
        layer_names=layer_names,
        layer_needles=[],
        function_names=function_names,
        function_needles=[],
        bundle_all=False,
    )


def _update_fuzzy_selection(
    kind: str,
    names: typing.List[str],
    selection: "definitions.Selection",
) -> "definitions.Selection":
    """Create a new selection object updated with the specified fuzzy match."""
    layer_needles = []
    if kind in ("*", "layer"):
        layer_needles = list(names)

    function_needles = []
    if kind in ("*", "function"):
        function_needles = list(names)

    return dataclasses.replace(
        selection,
        function_names=[],
        function_needles=function_needles,
        layer_names=[],
        layer_needles=layer_needles,
        bundle_all=False,
    )


def run(ex: "interactivity.Execution"):
    """Specify the target function(s) and/or layer(s) to target."""
    selection: "definitions.Selection" = ex.shell.selection
    is_exact = ex.args.get("exact", False)
    kind = ex.args["kind"]

    names = ex.args.get("name") or "*"
    if names and isinstance(names, str):
        names = [names]
    else:
        names = list(names)

    if kind == "*" and names == ["*"]:
        status = "ALL"
        message = "Selection has been cleared. All items are now selected."
        ex.shell.selection = dataclasses.replace(
            selection,
            function_needles=["*"],
            layer_needles=["*"],
            bundle_all=True,
        )
    elif is_exact:
        status = "EXACT"
        message = "Exact selection has been applied."
        ex.shell.selection = _update_exact_selection(kind, names, selection)
    else:
        status = "MATCH"
        message = "Matching items have been selected."
        ex.shell.selection = _update_fuzzy_selection(kind, names, selection)

    targets = ex.shell.context.get_selected_targets(ex.shell.selection)
    functions = [n for t in targets.function_targets for n in t.names]
    layers = [n for t in targets.layer_targets for n in t.names]
    return ex.finalize(
        status=status,
        message=message,
        echo=True,
        info={
            "functions": functions or "None",
            "layers": layers or "None",
        },
    )



from .abstracts import DataWrapper  # noqa
from .abstracts import Specification  # noqa
from .abstracts import Specification  # noqa
from .aws import AwsConnection  # noqa
from .configurations import Bundle  # noqa
from .configurations import Configuration  # noqa
from .configurations import Dependency  # noqa
from .configurations import EnvironmentVariable  # noqa
from .configurations import PipDependency  # noqa
from .configurations import PipperDependency  # noqa
from .configurations import PoetryDependency  # noqa
from .configurations import Target  # noqa
from .contexts import Context  # noqa
from .enumerations import DefaultFile  # noqa
from .enumerations import DependencyType  # noqa
from .enumerations import RUNTIME_VERSION  # noqa
from .enumerations import TargetType  # noqa
from .infomatics import LambdaFunction  # noqa
from .infomatics import LambdaLayer  # noqa
from .infomatics import LambdaLayerReference  # noqa
from .infomatics import PublishedLayer  # noqa
from .selections import Selection  # noqa



import datetime
import functools
import os
import typing

from reviser import definitions
from reviser import utils


def upload_callback(bytes_uploaded: int, status: dict):
    """Update the display with current update status"""
    status["uploaded_bytes"] += bytes_uploaded
    progress = int(100 * status["uploaded_bytes"] / status["total_bytes"])
    size = utils.to_human_readable_size(status["uploaded_bytes"])

    if progress - status["display_progress"] >= 5:
        print(f"{progress}% ({size})")
        status["display_progress"] = progress


def upload(
    target: "definitions.Target",
    dry_run: bool = False,
) -> typing.List[str]:
    """
    Upload deployment bundle using Boto3 with progress callback
    display

    :param target:
        The lambda function(s)/layer(s) target that will be uploaded
        for all selected names in the target.
    :param dry_run:
        If true, the upload process will be exercised without actually
        uploading the data.
    """
    try:
        size_bytes = os.path.getsize(target.bundle_zip_path)
    except FileNotFoundError:
        if not dry_run:
            raise RuntimeError("Bundling is required before deploy.")
        size_bytes = 0

    upload_status = {
        "uploaded_bytes": 0,
        "total_bytes": size_bytes,
        "display_progress": -5,
    }

    now = datetime.datetime.utcnow()
    timestamp = now.isoformat().replace(":", "-").replace(".", "-")
    s3_keys = [
        f"lambda-deployer/{target.kind.value}/{name}/{timestamp}.zip"
        for name in target.names
    ]

    size = utils.to_human_readable_size(upload_status["total_bytes"])
    print(f"[UPLOADING]: {s3_keys[0]} (size: {size})")

    client = target.client("s3")

    if not dry_run:
        client.upload_file(
            Filename=str(target.bundle_zip_path),
            Bucket=target.bucket,
            Key=s3_keys[0],
            Callback=functools.partial(upload_callback, status=upload_status),
        )

    for key in s3_keys[1:]:
        print(f"[COPYING]: {s3_keys[0]} -> {key}")
        if not dry_run:
            client.copy_object(
                Bucket=target.bucket,
                Key=key,
                CopySource={
                    "Bucket": target.bucket,
                    "Key": s3_keys[0],
                },
            )

    return s3_keys



import typing

import boto3
from botocore.client import BaseClient
from botocore.credentials import Credentials


class AwsConnection:
    """Boto3 Session wrapper that facilitates broader session management."""

    def __init__(self, session: boto3.Session = None):
        """..."""
        self._session = session or boto3.Session()
        self._aws_account = self.client("sts").get_caller_identity()
        self._aws_account_aliases = self.client("iam").list_account_aliases()

    @property
    def session(self) -> boto3.Session:
        """Returns the session object associated with this connection."""
        return self._session

    @property
    def user_id(self) -> str:
        """Returns the fully-qualified user ID for the associated session."""
        return self._aws_account.get("UserId")

    @property
    def user_arn(self) -> str:
        """Returns the ARN of the user associated with this session."""
        return self._aws_account.get("Arn")

    @property
    def user_slug(self) -> str:
        """
        Display slug from the user ARN containing account ID and user
        name or assumed role for use in shell prompt.
        """
        if "iam::" in self.user_arn:
            return self.user_arn.split("iam::")[-1]
        return self.user_arn.split("sts::")[-1].rsplit("/", 1)[0]

    @property
    def aws_account_id(self) -> str:
        """Returns the AWS account ID for the associated session."""
        return self._aws_account.get("Account")

    @property
    def aws_account_alias(self) -> typing.Optional[str]:
        """
        Returns the AWS account alias for the account associated with
        the session.
        """
        aliases = self._aws_account_aliases.get("AccountAliases")
        if aliases:
            return aliases[0]
        return None

    def get_credentials(self) -> Credentials:
        """Returns the credentials associated with the session."""
        return self.session.get_credentials()

    def client(self, *args, **kwargs) -> BaseClient:
        """Creates a new client of the specified type."""
        return self._session.client(*args, **kwargs)



import dataclasses
import datetime
import typing
from collections import defaultdict

from botocore.client import BaseClient

from reviser import definitions
from reviser import templating


def get_function_versions(
    lambda_client: BaseClient,
    function_name: str,
) -> typing.List["definitions.LambdaFunction"]:
    """Lists all of a function's versions and their associated aliases."""
    try:
        paginator = lambda_client.get_paginator("list_versions_by_function")
        versions = [
            definitions.LambdaFunction(item)
            for response in paginator.paginate(FunctionName=function_name)
            for item in response["Versions"]
        ]
    except lambda_client.exceptions.ResourceNotFoundException:
        return []

    aliases = lambda_client.list_aliases(FunctionName=function_name, MaxItems=500)
    aliased = defaultdict(list)
    for a in aliases["Aliases"]:
        aliased[a["FunctionVersion"]].append(a)

    versions.sort(
        key=lambda x: int(1e8 if x.version == "$LATEST" else (x.version or 0)),
    )
    return [
        dataclasses.replace(v, alias_responses=aliased[v.version]) for v in versions
    ]


def remove_function_version(lambda_client, version_arn: str):
    """
    Removes the lambda function version with the specified version ARN. This
    will fail for any versions.

    :param lambda_client:
        Boto3 client for interacting with the lambda function.
    :param version_arn:
        The ARN for the version of the lambda function that should be removed.
    """

    try:
        lambda_client.delete_function(FunctionName=version_arn)
        print("[PRUNED]: {}".format(version_arn))
    except Exception as error:
        print(
            '[FAILED]: Version {} could not be deleted "{}"'.format(version_arn, error)
        )


def echo_function_versions(client: BaseClient, names: typing.List[str]):
    """
    Displays function versions with basic info for each specified
    lambda function.
    """
    for name in names:
        versions = get_function_versions(client, name)
        if not versions:
            print(f'\n[IGNORED]: No function "{name}" was found.')
            continue

        groups = defaultdict(list)
        for v in versions:
            dt = datetime.datetime.fromisoformat((v.modified or "").split("+")[0])
            groups[str(dt.date())].append({"time": dt.strftime("%H:%M"), "data": v})

        templating.printer(
            "servicer/functioning/echo_versions.jinja2",
            name=name,
            groups=list(sorted(groups.items(), key=lambda x: x[0])),
        )


def get_function_version(
    lambda_client: BaseClient,
    function_name: str,
    qualifier: str = "$LATEST",
) -> "definitions.LambdaFunction":
    """Retrieves the configuration for the specified lambda function."""
    response = lambda_client.get_function_configuration(
        FunctionName=function_name,
        Qualifier=qualifier or "$LATEST",
    )

    request = {
        "FunctionName": function_name,
        "MaxItems": 500,
    }
    if (version := response.get("version")) and version != "$LATEST":
        request["FunctionVersion"] = version
    alias_response = lambda_client.list_aliases(**request)

    return definitions.LambdaFunction(
        response,
        alias_response.get("Aliases") or [],
    )



import argparse
import os
import pathlib
import sys

import reviser


def _suppress(condition: bool, help_doc: str) -> str:
    """Returns the help status based on the suppression condition."""
    if condition:
        return argparse.SUPPRESS
    return help_doc


def _get_auth_directory() -> pathlib.Path:
    """Finds the first valid and specified AWS credentials directory."""
    if creds_path := os.environ.get("AWS_SHARED_CREDENTIALS_FILE"):
        return pathlib.Path(creds_path).expanduser().parent.absolute()

    if config_path := os.environ.get("AWS_CONFIG_FILE"):
        return pathlib.Path(config_path).expanduser().parent.absolute()

    return pathlib.Path("~/.aws").expanduser().absolute()


def create_parser(internal_parser: bool) -> argparse.ArgumentParser:
    """..."""
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        prog="reviser",
    )

    # These are arguments from the wrapping command line interface that
    # may bleed through into the execution and should be ignored here
    # but not lead to a parsing error for unknown arguments. These also
    # serve to prevent conflicting arguments from being created here in
    # the future as these are reserved for the wrapping command execution.
    info = sys.version_info
    parser.add_argument(
        "--runtime",
        default=f"{info.major}.{info.minor}",
        help=_suppress(
            internal_parser,
            """
            Python runtime version of the lambda function(s)/layer(s)
            that will be deployed to from this shell.
            """,
        ),
    )

    parser.add_argument(
        "--tag-version",
        dest="image_tag_version",
        default=reviser.__version__,
        help=_suppress(
            internal_parser,
            """
            Image tag version prefix to use to launch the container. Defsults
            to the version of reviser used to launch the container.
            """,
        ),
    )

    parser.add_argument(
        "--aws-directory",
        default=str(_get_auth_directory()),
        help=_suppress(
            internal_parser,
            """
            Specifies the directory where AWS credentials are stored.
            This directory will be mounted as a volume into the shell
            container at launch to make credentials available within
            the running container for interacting with AWS resources.
            """,
        ),
    )

    parser.add_argument(
        "-d",
        "--directory",
        dest="root_directory",
        default=os.path.realpath(os.curdir),
        help="""
            The root directory where the lambda shell executions will
            take place. By default it is the current directory at the
            time the shell is started.
            """,
    )
    parser.add_argument(
        "-p",
        "--profile",
        dest="aws_profile_name",
        help="""
            The profile name for the AWS session that will be used to
            interact with lambda during the deployment process. If not
            set the default profile will be used if environment variables
            are not set as an overriding alternative.
            """,
    )
    parser.add_argument(
        "--run",
        dest="command_group_name",
        help="""
            Specifies a run command group to execute in a new shell
            session that exits after the commands have been executed.
            These command groups are useful for defining common operations
            to execute in a single command without opening a persistent
            shell. They are also useful for CI purposes. The value here
            must be specified in the lambda.yaml file under the top-level
            run key and must be a list of shell commands to execute.
            """,
    )

    return parser



import typing
import yaml

from botocore.client import BaseClient

from reviser import definitions
from reviser import servicer


def _get_layer_updates(
    client: BaseClient,
    function_name: str,
    target: "definitions.Target",
    current_configuration: dict,
    published_layers: typing.List["definitions.PublishedLayer"],
) -> typing.Optional[typing.List[str]]:
    """
    Creates an updated list of layer ARNs to update the function layer
    configuration to during function configuration update.
    """
    if target.ignores_any("layer", "layers"):
        return None

    existing = [item["Arn"] for item in (current_configuration.get("Layers") or [])]
    updates = {p.arn: p.versionedArn for p in published_layers}
    layers = [
        updates.get(a.arn, a.arn)
        for a in target.layer_attachments
        if a.is_attachable(function_name)
    ]

    latest = []
    for arn in layers:
        if arn and arn.count(":") == 7:
            latest.append(arn)
        elif arn and (found_versions := servicer.get_layer_versions(client, arn)):
            version = found_versions[-1].version
            latest.append(f"{arn}:{version}")
        else:
            print(f"\n[WARNING]: No versions found for layer:\n   {arn}\n")

    if latest == existing:
        return None
    return latest


def _get_handler_update(
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[str]:
    """Specifies an updated handler if a change is found."""
    latest = target.bundle.handler
    existing = current_configuration.get("Handler")
    if target.ignores_any("handler") or latest == existing:
        return None
    return latest


def _get_runtime_update(
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[str]:
    """Specifies an updated runtime if a change is found."""
    latest = f"python{definitions.RUNTIME_VERSION}"
    existing = current_configuration.get("Runtime")
    if target.ignores_any("runtime") or latest == existing:
        return None
    return latest


def _get_memory_update(
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[int]:
    existing = current_configuration.get("MemorySize")
    latest = target.memory
    if target.ignores_any("memory") or latest == existing:
        return None
    return latest


def _get_timeout_update(
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[int]:
    existing = current_configuration.get("Timeout")
    latest = target.timeout
    if target.ignores_any("timeout") or latest == existing:
        return None
    return latest


def _get_variable_updates(
    function_name: str,
    target: "definitions.Target",
    current_configuration: dict,
) -> typing.Optional[dict]:
    """
    Determines the environment variable updates for the function.
    """
    if target.ignores_any("variable", "variables"):
        return None

    existing = current_configuration.get("Environment", {}).get("Variables", {})

    def _get_value(variable: "definitions.EnvironmentVariable"):
        """Fetch the environment variable value based on its settings."""
        if variable.preserve:
            return existing.get(variable.name)
        return variable.get_value(function_name)

    latest = {e.name: v for e in target.variables if (v := _get_value(e))}
    if latest == existing:
        return None
    return {"Variables": latest}


def update_function_configuration(
    function_name: str,
    target: "definitions.Target",
    published_layers: typing.List["definitions.PublishedLayer"],
    dry_run: bool = False,
):
    """Updates the function configuration as part of a deployment process."""
    client = target.client("lambda")
    current = client.get_function_configuration(FunctionName=function_name)

    changes = dict(
        Layers=_get_layer_updates(
            client,
            function_name,
            target,
            current_configuration=current,
            published_layers=published_layers,
        ),
        Runtime=_get_runtime_update(target, current),
        MemorySize=_get_memory_update(target, current),
        Timeout=_get_timeout_update(target, current),
        Environment=_get_variable_updates(function_name, target, current),
        Handler=_get_handler_update(target, current),
    )

    modifications = {k: v for k, v in changes.items() if v is not None}
    if not modifications:
        print(f"[UNCHANGED]: {function_name} configuration unchanged.")
        return

    print(f"[MODIFYING]: Updating {function_name} configuration values:")
    print(yaml.safe_dump(modifications))

    if dry_run:
        print("[DRY_RUN]: No configuration updates applied.")
        return

    client.update_function_configuration(FunctionName=function_name, **modifications)



import pathlib as _pathlib
from importlib import metadata as _metadata

try:
    from reviser.interactivity import run_shell  # noqa
except ImportError as error:
    _stored_error = error

    def run_shell(*args, **kwargs):  # type:ignore
        """Replacement shell when import fails due to missing dependencies."""
        global _stored_error
        raise _stored_error


try:
    __version__ = _metadata.version(__package__)
except _metadata.PackageNotFoundError:  # pragma: no-cover
    # If the package is not installed such that it has distribution metadata
    # fallback to loading the version from the pyproject.toml file.
    import toml as _toml

    __version__ = _toml.loads(
        _pathlib.Path(__file__).parent.parent.joinpath("pyproject.toml").read_text()
    )["tool"]["poetry"]["version"]



"""
Removes old function and/or layer versions for the selected targets.
"""
import textwrap
import argparse
import typing

from botocore.client import BaseClient

from reviser import definitions
from reviser import interactivity
from reviser import servicer


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Shell auto-completes for this command."""
    return ["--start", "--end", "--dry-run"]


def populate_subparser(parser: argparse.ArgumentParser):
    """Add subcommand options for this command."""
    parser.add_argument(
        "--start",
        type=int,
        default=None,
        help="Keep versions lower (earlier/before) this one.",
    )

    parser.add_argument(
        "--end",
        type=int,
        default=None,
        help="Do not prune versions higher than this value.",
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Echo pruning operation without actually executing it.",
    )

    parser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        help="Run the prune process without reviewing first.",
    )


def _prune_function(
    lambda_client: BaseClient,
    function_name: str,
    start: int = None,
    end: int = None,
    dry_run: typing.Optional[bool] = False,
    confirm: typing.Optional[bool] = True,
):
    """
    Executes a pruning operation on the given lambda functions.

    :param lambda_client:
        Boto3 client interface for the lambda service
    :param function_name:
        Name of the function to prune
    :param start:
        First version of the lambda function to prune. Versions below
        that will be retained.
    :param end:
        Inclusive end version of the lambda function to prune. Versions
        above that will be retained.
    :param dry_run:
        If true, only print out what would happen and don't actually prune
        the versions.
    :param confirm:
        Whether or not to ask before proceeding with the prune operation.
    """
    versions = servicer.get_function_versions(lambda_client, function_name)
    removal_arns = [
        v.arn
        for v in versions
        if v.version != "$LATEST"
        and v.arn
        and (start is None or start <= int(v.version or 0))
        and (end is None or int(v.version or 0) <= end)
        and not v.aliases
    ]

    print("\nARN Versions to be removed:")
    print(
        textwrap.indent(
            "\n".join(removal_arns),
            prefix="  - ",
        )
    )

    if dry_run:
        print("\n[DRY RUN]: Skipped removal process.")
        return

    abort = confirm and not (
        input("\nExecute prune action [y/N]? ") or ""
    ).lower().startswith("y")
    if abort:
        print("\n[ABORTED]: Skipped removal process.")
        return

    for arn in removal_arns:
        servicer.remove_function_version(lambda_client, arn)

    return removal_arns


def _prune_layer(
    lambda_client: BaseClient,
    layer_name: str,
    start: int = None,
    end: int = None,
    dry_run: typing.Optional[bool] = False,
    confirm: typing.Optional[bool] = True,
):
    """
    Executes a pruning operation on the given lambda layers.

    :param lambda_client:
        Boto3 client interface for the lambda service.
    :param layer_name:
        Name of the layer to prune
    :param start:
        First version of the lambda layer to prune. Versions below
        that will be retained.
    :param end:
        Inclusive end version of the lambda layer to prune. Versions
        above that will be retained.
    :param dry_run:
        If true, only print out what would happen and don't actually prune
        the versions.
    :param confirm:
        Whether or not to ask before proceeding with the prune operation.
    """
    versions = servicer.get_layer_versions(lambda_client, layer_name)
    removals = [
        version
        for version in versions[:-1]
        if (start is None or start <= int(version.version or 0))
        and (end is None or int(version.version or 0) <= end)
    ]
    arns = [r.arn for r in removals if r.arn]

    print("\nARN Versions to be removed:")
    print(
        textwrap.indent(
            "\n".join(arns or []),
            prefix="  - ",
        )
    )

    if dry_run:
        print("\n[DRY RUN]: Skipped removal process.")
        return

    abort = confirm and not (
        input("\nExecute prune action [y/N]? ") or ""
    ).lower().startswith("y")
    if abort:
        print("\n[ABORTED]: Skipped removal process.")
        return

    for arn in arns:
        print(f"[PRUNING]: {arn}")
        servicer.remove_layer_version(lambda_client, arn)

    return arns


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Runs the pruning operation on the targets."""
    selected = ex.shell.context.get_selected_targets(ex.shell.selection)
    targets = sorted(selected.targets, key=lambda t: t.kind.value)

    caller = {
        definitions.TargetType.FUNCTION: _prune_function,
        definitions.TargetType.LAYER: _prune_layer,
    }

    results = {
        name: caller[target.kind](
            target.client("lambda"),
            name,
            start=ex.args.get("start"),
            end=ex.args.get("end"),
            dry_run=ex.args.get("dry_run") or False,
            confirm=not ex.args.get("yes") or False,
        )
        for target in targets
        for name in target.names
    }

    return ex.finalize(
        status="PRUNED",
        message="Specified versions have been removed.",
        info={name: removed_arns for name, removed_arns in results.items()},
        echo=True,
    )



import dataclasses
import json
import pathlib
import subprocess
import sys
import typing

from reviser.definitions import abstracts
from reviser.definitions import configurations
from reviser.definitions import enumerations


@dataclasses.dataclass(frozen=True)
class Dependency(abstracts.Specification):
    """Package dependency data structure."""

    target: "configurations.Target"

    @property
    def kind(self) -> "enumerations.DependencyType":
        """Kind of dependency"""
        return enumerations.DependencyType(
            value=self.get(
                "kind",
                default=enumerations.DependencyType.PIP.value,
            )
        )

    @property
    def file(self) -> typing.Optional[pathlib.Path]:
        """Optionally specified path used to load dependencies."""
        if value := self.get("file"):
            return self.directory.joinpath(value).absolute()

        if self.get("packages") is not None:
            return None

        default = self.directory.joinpath(
            enumerations.DefaultFile.from_dependency_type(self.kind).value
        )
        return default if default.exists() else None

    @property
    def packages(self) -> typing.List[str]:
        """Explicitly defined packages."""
        return self.get_first_as_list(["packages"], ["package"], default=[])

    def get_package_names(self) -> typing.List[str]:
        """
        Returns a list of package names to install collected from
        the various ways packages can be specified.
        """
        return []


@dataclasses.dataclass(frozen=True)
class PipDependency(Dependency):
    """Pip package dependency data structure."""

    def get_package_names(self) -> typing.List[str]:
        """
        Returns a list of package names to install collected from
        the various ways packages can be specified.
        """
        packages = self.packages.copy()
        if self.file:
            packages += [
                item
                for line in (self.file.read_text() or "").split("\n")
                if (item := line.strip())
            ]
        return packages

    def serialize(self) -> dict:
        """Serializes the object for output representation."""
        return {
            "kind": self.kind.value,
            "packages": self.get_package_names(),
        }


@dataclasses.dataclass(frozen=True)
class PipperDependency(Dependency):
    """Pipper package dependency data structure."""

    @property
    def prefix(self) -> typing.Optional[str]:
        """Custom S3 key prefix in the bucket where the repository resides."""
        return self.get("prefix")

    @property
    def bucket(self) -> typing.Optional[str]:
        """S3 bucket where the pipper repository resides."""
        buckets = self.get("buckets", default=self.get("bucket", default=None))

        if not buckets and self.file:
            return (self.get_package_data() or {}).get("bucket") or None

        if isinstance(buckets, str):
            return buckets

        return buckets[self.connection.aws_account_id]

    def get_package_data(self) -> typing.Optional[dict]:
        """Returns the data stored in the specified package file."""
        if self.file:
            return json.loads(self.file.read_text())
        return None

    def get_package_names(self) -> typing.List[str]:
        """
        Returns a list of package names to install collected from
        the various ways packages can be specified.
        """
        packages = self.packages.copy()
        if data := self.get_package_data():
            packages += data.get("dependencies") or []
        return packages

    def serialize(self) -> dict:
        """Serializes the object for output representation."""
        return {
            "kind": self.kind.value,
            "prefix": self.prefix,
            "bucket": self.bucket,
            "packages": self.get_package_names(),
        }


@dataclasses.dataclass(frozen=True)
class PoetryDependency(Dependency):
    """Poetry package dependency data structure."""

    @property
    def extras(self) -> typing.List[str]:
        """List of extra packages to install."""
        return self.get_as_list("extras", default=self.get_as_list("extra")) or []

    def get_package_names(self) -> typing.List[str]:
        """
        Returns a list of package names to install collected from
        the various ways packages can be specified.
        """
        packages = self.packages.copy()
        if not self.file:
            return packages

        directories = [
            pathlib.Path(sys.prefix).joinpath("bin"),
            pathlib.Path(sys.executable).parent,
        ]
        directories = [d for d in directories if d.exists()]

        finder = (
            str(p)
            for d in directories
            for p in d.iterdir()
            if p.name == "poetry" or p.name.startswith("poetry.")
        )
        executable = next(finder, None)
        if executable is None:
            raise RuntimeError(
                "Unable to find poetry installation in current Python environment."
            )

        command = [
            executable,
            "export",
            "--format=requirements.txt",
            "--without-hashes",
        ]

        for group in self.extras:
            command.append(f"--extras={group}")

        result = subprocess.run(command, stdout=subprocess.PIPE, check=True)

        packages += [
            item
            for line in result.stdout.decode().strip().split("\n")
            if (item := line.split(";")[0].strip())
        ]

        return packages

    def serialize(self) -> dict:
        """Serializes the object for output representation."""
        return {
            "kind": self.kind.value,
            "packages": self.get_package_names(),
        }



import dataclasses
import typing
import fnmatch

from reviser.definitions import abstracts
from reviser.definitions import configurations


@dataclasses.dataclass(frozen=True)
class EnvironmentVariable(abstracts.Specification):
    """Data structure for function environment variables."""

    target: "configurations.Target"

    @property
    def name(self) -> str:
        """Name for the environment variable."""
        if (name := self.get("name")) is not None:
            return name

        if (arg := self.get("arg")) is not None:
            return arg.split("=")[0]

        return "unknown-environment-variable"

    @property
    def preserve(self) -> bool:
        """
        Specifies whether or not the environment variable should be
        preserved, such that changes or settings here will not be
        updated. Useful for variables that are managed or assigned by
        other systems.
        """
        return self.get("preserve", default=False)

    @property
    def restrictions(self) -> typing.List[str]:
        """Only attach to functions matching values in this list."""
        value = self.get("only", default=[])
        if isinstance(value, str):
            return [value]
        return value

    @property
    def exclusions(self) -> typing.List[str]:
        """Don't attach to functions matching values in this list."""
        value = self.get("except", default=[])
        if isinstance(value, str):
            return [value]
        return value

    def get_value(self, function_name: str) -> typing.Optional[str]:
        """
        Retrieves the value of the environment variable for the given
        function name.
        """
        if not self.has("value") and (arg := self.get("arg")):
            value = arg.split("=", 1)[1]
        else:
            value = self.get("value")

        if isinstance(value, dict):
            selector = (
                v
                for pattern, v in value.items()
                if fnmatch.fnmatch(function_name, pattern)
            )
            value = next(selector, None)

        includer = (
            True
            for pattern in self.restrictions
            if fnmatch.fnmatch(function_name, pattern)
        )
        if not next(includer, not self.restrictions):
            return None

        excluder = (value for e in self.exclusions if fnmatch.fnmatch(function_name, e))
        return next(excluder, value)

    def serialize(self) -> dict:
        if self.preserve:
            kwargs = {}
        else:
            kwargs = {
                "values": {
                    n: v
                    for n in self.target.names
                    if (v := self.get_value(n)) is not None
                },
            }

        return {
            "name": self.name,
            "only": self.restrictions,
            "except": self.exclusions,
            "preserve": self.preserve,
            **kwargs,
        }



import os
import pathlib
import subprocess
import sys
import typing

from reviser import parsing


def _remove_arg(
    source: typing.List[str],
    flag: str,
    has_value: bool,
) -> typing.List[str]:
    """
    Searches for the specified flag within the source arguments and removes it
    and its optional value if has_value is specified, returning a copy of the source
    list with those elements removed. If the flag is not found an unaltered copy of
    the source list is returned instead. This is used in the CLI command to prevent
    certain flags from being used in both the host and container invocation contexts.
    """
    try:
        index = source.index(flag)
        offset = 1 if has_value else 0
        next_index = index + 1 + offset
        return source[:index] + source[next_index:]
    except ValueError:
        pass

    if not flag.startswith("--"):
        return source.copy()

    try:
        finder = (i for i, x in enumerate(source) if x.startswith(f"{flag}="))
        index = next(finder)
        next_index = index + 1
        return source[:index] + source[next_index:]
    except StopIteration:
        pass

    return source.copy()


def main(arguments: typing.List[str] = None) -> bool:
    """Start the shell container based on the command line arguments."""
    print(f"\n[DIRECTORY]: {os.path.realpath(os.curdir)}")
    args = parsing.create_parser(False).parse_args(args=arguments)

    credentials_directory = pathlib.Path(args.aws_directory).absolute()
    deploy_directory = pathlib.Path(args.root_directory).absolute()
    folder_name = deploy_directory.name
    version = args.image_tag_version

    container_args = [a.replace("\\", "/") for a in (arguments or sys.argv[1:])]
    container_args = _remove_arg(container_args, "-d", True)
    container_args = _remove_arg(container_args, "--directory", True)
    container_args = _remove_arg(container_args, "--tag-version", True)

    command = [
        "docker",
        "run",
        "-it",
        "--rm",
        "-v",
        "{}:/root/.aws".format(str(credentials_directory).replace("\\", "/")),
        "-v",
        "{}:/project/{}".format(str(deploy_directory).replace("\\", "/"), folder_name),
        "--workdir=/project/{}".format(folder_name),
        f"swernst/reviser:{version}-{args.runtime}",
        *container_args,
    ]

    display = " ".join(command).replace(" -", "\n        -")
    print(f"[RUN]: {display}\n")

    try:
        subprocess.check_call(command)
        return True
    except subprocess.CalledProcessError:
        return False


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reviser',
 'reviser.bundling',
 'reviser.commands',
 'reviser.commands.helper',
 'reviser.definitions',
 'reviser.definitions.configurations',
 'reviser.deploying',
 'reviser.interactivity',
 'reviser.interactivity.shells',
 'reviser.servicer',
 'reviser.servicer.functioning',
 'reviser.templating']

package_data = \
{'': ['*']}

extras_require = \
{'shell': ['boto3>=1.16.11,<2.0.0',
           'colorama>=0.4.4,<0.5.0',
           'Jinja2>=2.11.2,<3.0.0',
           'prompt-toolkit>=3.0.8,<4.0.0',
           'PyYAML>=5.3.1,<6.0.0',
           'pipper>=0.7.0,<0.8.0',
           'toml>=0.10.2,<0.11.0',
           'poetry>=1.1.4,<2.0.0']}

entry_points = \
{'console_scripts': ['reviser = reviser.cli:main']}

setup_kwargs = {
    'name': 'reviser',
    'version': '0.1.8',
    'description': 'AWS Lambda function/layer version deployment manager.',
    'long_description': '# Reviser (v0.1.8)\n\n[![PyPI version](https://badge.fury.io/py/reviser.svg)](https://pypi.org/project/reviser/)\n[![build status](https://gitlab.com/rocket-boosters/reviser/badges/main/pipeline.svg)](https://gitlab.com/rocket-boosters/reviser/commits/main)\n[![coverage report](https://gitlab.com/rocket-boosters/reviser/badges/main/coverage.svg)](https://gitlab.com/rocket-boosters/reviser/commits/main)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Code style: flake8](https://img.shields.io/badge/code%20style-flake8-white)](https://gitlab.com/pycqa/flake8)\n[![Code style: mypy](https://img.shields.io/badge/code%20style-mypy-white)](http://mypy-lang.org/)\n[![PyPI - License](https://img.shields.io/pypi/l/reviser)](https://pypi.org/project/reviser/)\n\nReviser is a tool for AWS Lambda function and layer version deployment and\nalias management specifically for Python runtimes where the actual\ninfrastructure is managed separately, mostly likely by CloudFormation or\nTerraform. There are a number of ways to manage AWS Lambda functions and layers\nalready, but their generality and all-encompassing approaches don\'t integrate\nwell with certain workflows and can be overly complex for many needs.\n\nReviser is scoped to facilitate the deployment and updating of AWS Lambda\nPython functions and layers for all of the version-specific configurations,\ne.g. code bundles, environment variables, memory size, and timeout lengths.\nThe expectation is that functions are created by other means and then\nconfiguration for versions is managed with the reviser through an interactive\nor scripted shell of commands.\n\n- [Basic Usage](#basic-usage)\n- [Shell Commands](#shell-commands)\n   - [alias](#alias)\n   - [bundle](#bundle)\n   - [configs](#configs)\n   - [deploy](#deploy)\n   - [exit](#exit)\n   - [help (?)](#help-)\n   - [list](#list)\n   - [prune](#prune)\n   - [push](#push)\n   - [reload](#reload)\n   - [select](#select)\n   - [shell](#shell)\n   - [status](#status)\n- [Configuration Files](#configuration-files)\n   - [bucket(s)](#buckets)\n   - [region](#region)\n   - [targets](#targets)\n      - [targets[N].kind](#targetsnkind)\n      - [targets[N].name(s)](#targetsnnames)\n      - [targets[N].region](#targetsnregion)\n      - [targets[N].dependencies](#targetsndependencies)\n      - [targets[N].dependencies(kind="pipper")](#targetsndependencieskindpipper)\n      - [targets[N].dependencies(kind="poetry")](#targetsndependencieskindpoetry)\n      - [targets[N].bundle](#targetsnbundle)\n         - [targets[N].bundle.include(s)](#targetsnbundleincludes)\n         - [targets[N].bundle.exclude(s)](#targetsnbundleexcludes)\n         - [targets[N].bundle.omit_package(s)](#targetsnbundleomit_packages)\n         - [targets[N].bundle.handler](#targetsnbundlehandler)\n   - [function targets](#function-targets)\n      - [(function) targets[N].layer(s)](#function-targetsnlayers)\n      - [(function) targets[N].memory](#function-targetsnmemory)\n      - [(function) targets[N].timeout](#function-targetsntimeout)\n      - [(function) targets[N].variable(s)](#function-targetsnvariables)\n      - [(function) targets[N].ignore(s)](#function-targetsnignores)\n   - [run](#run)\n\n# Basic Usage\n\nA project defines one or more lambda function configuration targets in a \n`lambda.yaml` file in the root project directory. The most basic configuration\nlooks like this:\n\n```yaml\nbucket: name-of-s3-bucket-for-code-uploads\ntargets:\n- kind: function\n  name: foo-function\n```\n\nThis configuration defines a single `foo-function` lambda function target that\nwill be managed by reviser. The expectation is that this function exists and\nwas created by another means, e.g. CloudFormation or Terraform. A bucket must\nbe specified to indicate where the zipped code bundles will be uploaded prior\nto them being applied to the target(s). The bucket must already exist as well.\n\nBy default the package will include no external, e.g. pip, package\ndependencies. It will search for the first folder in the directory where the\n`lambda.yaml` file is located that contains an `__init__.py` file, identifying\nthat folder as a Python source package for the function. It will also look for\na `lambda_function.py` alongside the `lambda.yaml` file to serve as the \nentrypoint. These will be included in the uploaded and deployed code bundle\nwhen a `push` or a `deploy` command is executed. These default settings can\nall be configured along with many more as will be outlined below.\n\nTo deploy this example project, install the reviser python library and\nstart the shell with the command `reviser` in your terminal of choice\nin the directory where the `lambda.yaml` file resides. Docker must be running\nand available in the terminal in which you execute this command, as reviser\nis a containerized shell environment that runs within a container that mimics\nthe actual AWS Lambda runtime environment. Then run the `push` command within\nthe launched shell to create and upload the bundled source code and publish\na new version of the `foo-function` lambda function with the uploaded results.\n\n# Shell commands\n\nThe reviser command starts an interactive shell within a Docker container \ncompatible with the AWS Python Lambda runtime. This shell contains various\ncommands for deploying and managing deployments of lambda functions and layers\ndefined in a project\'s `lambda.yaml` configuration file, the format of which\nis described later in this document. The shell commands are:\n\n## alias\n\nAssigns an alias to the specified version of the selected or specified lambda\nfunction. Or it will create a new alias and assign it to the specified version\nif the --create flag is included.\n\n```\nusage: alias [--function FUNCTION] [--yes] [--create] alias version\n\npositional arguments:\n  alias                Name of an existing alias to move to the specified\n                       version, or the name of an alias to create and assign\n                       to the specified function version if the --create flag\n                       is included to allow for creating a new alias.\n  version              Version of the function that the alias should be\n                       assigned to. This will either be an integer value or\n                       $LATEST. To see what versions are available for a given\n                       function use the list command.\n\noptional arguments:\n  --function FUNCTION  The alias command only acts on one function. This can\n                       be achieved either by selecting the function target via\n                       the select command, or specifying the function name to\n                       apply this change to with this flag.\n  --yes                By default this command will require input confirmation\n                       before carrying out the change. Specify this flag to\n                       skip input confirmation and proceed without a breaking\n                       prompt.\n  --create             When specified the alias will be created instead of\n                       reassigned. Use this to create and assign new aliases\n                       to a function. When this flag is not specified, the\n                       command will fail if the alias doesn\'t exist, which\n                       helps prevent accidental alias creation.\n\n```\n\nTo assign an existing `test` alias to version 42 of the selected\nfunction, the command would be:\n\n```\n> alias test 42\n```\n\nIf multiple functions are currently selected, use `--function=<NAME>`\nto identify the function to which the alias change will be applied.\n\n## bundle\n\nInstalls dependencies and copies includes into a zipped file that is structured\ncorrectly to be deployed to the lambda function/layer target.\n\n```\nusage: bundle [--reinstall]\n\noptional arguments:\n  --reinstall  Add this flag to reinstall dependencies on a repeated bundle\n               operation. By default, dependencies will remain cached for the\n               lifetime of the shell to speed up the bundling process. This\n               will force dependencies to be installed even if they had been\n               installed previously.\n\n```\n\n## configs\n\nDisplays the configs loaded from the lambda.yaml file and fully populated with\ndefaults and dynamic values. Use this to inspect and validate that the loaded\nconfiguration meets expectations when parsed into the reviser shell.\n\n```\nusage: configs\n\n```\n\n## deploy\n\nUploads the bundled contents to the upload S3 bucket and then publishes a new\nversion of each of the lambda targets with that new bundle and any modified\nsettings between the current configuration and that target\'s existing\nconfiguration. This command will fail if a target being deployed has not\nalready been bundled.\n\n```\nusage: deploy [--description DESCRIPTION] [--dry-run]\n\noptional arguments:\n  --description DESCRIPTION\n                        Specify a message to assign to the version published\n                        by the deploy command.\n  --dry-run             If set, the deploy operation will be exercised without\n                        actually carrying out the actions. This can be useful\n                        to validate the deploy process without side effects.\n\n```\n\n## exit\n\nExits the shell and returns to the parent terminal.\n\n```\nusage: exit\n\n```\n\n## help (?)\n\nDisplays help information on the commands available within the shell.\nAdditional help on each command can be found using the --help flag on the\ncommand in question.\n\n```\nusage: help\n\n```\n\n## list\n\nLists versions of the specified lambda targets with info about each version.\n\n```\nusage: list\n\n```\n\n## prune\n\nRemoves old function and/or layer versions for the selected targets.\n\n```\nusage: prune [--start START] [--end END] [--dry-run] [-y]\n\noptional arguments:\n  --start START  Keep versions lower (earlier/before) this one.\n  --end END      Do not prune versions higher than this value.\n  --dry-run      Echo pruning operation without actually executing it.\n  -y, --yes      Run the prune process without reviewing first.\n\n```\n\n## push\n\nCombined single command for bundling and deploying the selected targets.\n\n```\nusage: push [--reinstall] [--description DESCRIPTION] [--dry-run]\n\noptional arguments:\n  --reinstall           Add this flag to reinstall dependencies on a repeated\n                        bundle operation. By default, dependencies will remain\n                        cached for the lifetime of the shell to speed up the\n                        bundling process. This will force dependencies to be\n                        installed even if they had been installed previously.\n  --description DESCRIPTION\n                        Specify a message to assign to the version published\n                        by the deploy command.\n  --dry-run             If set, the deploy operation will be exercised without\n                        actually carrying out the actions. This can be useful\n                        to validate the deploy process without side effects.\n\n```\n\n## reload\n\nReloads the lambda.yaml configuration file from disk.\n\n```\nusage: reload\n\n```\n\n## select\n\nAllows for selecting subsets of the targets within the loaded configuration.\nThe subsets are fuzzy-matched unless the --exact flag is used.\n\n```\nusage: select [--exact] {function,layer,*} [name]\n\npositional arguments:\n  {function,layer,*}  Specify the type of objects to select or use * to select\n                      both types.\n  name                Specifies the value to match against the function and\n                      layer target names available from the configuration.\n                      This can include shell-style wildcards and will also\n                      match against partial strings. If the --exact flag is\n                      specified, this value must exactly match one of the\n                      targets instead of the default fuzzy matching behavior.\n\noptional arguments:\n  --exact             Forces the match to be exact instead of fuzzy.\n\n```\n\n## shell\n\nSpecial command to use in run command groups/macros to start interactive\ncommand mode for the terminal. Useful when in scenarios where you wish to\nprefix an interactive session with commonly executed commands. For example, if\nyou want to select certain targets with the select command as part of starting\nthe shell, you could create a run command group/macro in your lambda.yaml that\nexecutes the select command and then executes the shell command. This would\nupdated the selection and then with the shell command, start the shell in\ninteractive mode. Without specifying the shell command here, the run command\ngroup/macro would just set a selection and then exit.\n\n```\nusage: shell\n\n```\n\n## status\n\nShows the current status information for each of the selected lambda targets.\n\n```\nusage: status [qualifier]\n\npositional arguments:\n  qualifier  Specifies a version or alias to show status for. If not\n             specified, $LATEST will be used for functions and the latest\n             version will be dynamically determined for layers.\n\n```\n    \nMore detail on any of these commands can be found from within the shell by\nexecuting them with the `--help` flag.\n\nThe reviser application also supports non-interactive batch command\nexecution via `run` macros that behave similarly to how `npm run <command>` \ncommands are defined. For more details see the `run` attribute section of the\nconfiguration file definitions below.\n\n# Configuration Files\n\nConfiguration files, named `lambda.yaml` define the the lambda targets to be\nmanaged within a project. The top-level keys in the configuration file are:\n\n## bucket(s)\n\nThis key defines the bucket or buckets where zipped source bundles will be\nuploaded before they are deployed to their lambda function and/or layer\ntargets. Basic usage is to specify the bucket as a key:\n\n```yaml\nbucket: bucket-name\n```\n\nIt\'s also possible for multi-account scenarios to specify multiple buckets as\na key-value pairing where the keys are the AWS account IDs (as strings) and\nthe values are the bucket names associated with those IDs. In this case the\nbucket selection is made dynamically based on the AWS session loaded during\nshell initialization. Specifying multiple buckets looks like:\n\n```yaml\nbuckets:\n  "123456789": bucket-in-account-123456789\n  "987654321": bucket-in-account-987654321\n``` \n\n## region\n\nThe AWS region in which teh resources reside can be specified at the top\nlevel of the file if desired. It is recommended that the region be specified\nwithin the calling AWS profile if possible for flexibility, but there are\nsituations where it makes more sense to make it explicit within the \nconfiguration file instead. If no region is found either in the configuration\nfile or in the AWS profile the `us-east-1` region will be used as the default\nin keeping with AWS region defaulting conventions. Specify the region with\nthe top-level key:\n\n```yaml\nregion: us-east-2\n```\n\n## targets\n\nTargets is where the bulk of the configuration resides. Each item\nis either of the *function* or *layer* kind and has associated\nconfiguration and bundling settings according to the type. Common\nto both *function* and *layer* kinds are the keys:\n\n### targets[N].kind\n\nAs mentioned already, each target must specify its object type using\nthe kind key:\n\n```yaml\ntargets:\n- kind: function\n  ...\n- kind: layer\n  ...\n```\n\n### targets[N].name(s)\n\nThe name specifies the name of the target object, not the ARN. For example,\na function named foo would be represented as:\n\n```yaml\ntargets:```  \n- kind: function\n  name: foo\n```\n\nA single target can point to multiple functions. This is useful in cases\nwhere a single target could be for both development and production functions\nor where a single code-base is shared across multiple functions for logical\nor integration reasons. In this case a list of names is supplied instead:\n\n```yaml\ntargets:\n- kind: function\n  names:\n  - foo-devel\n  - foo-prod\n``` \n\n### targets[N].region\n\nIn the same fashion as regions can be explicitly set as a top-level\nconfiguration key, they can also be set on a per-target basis. If set,\nthe target region will take precedence over the top-level value and\nthe profile-specified value. This makes deploying code across regions\nwithin a single configuration file possible.\n\n### targets[N].dependencies\n\nDependencies is a list of external dependency sources to install as\nsite packages in the lambda function or layer. Multiple package managers\nare supported and specified by the `kind` attribute:\n\n```yaml\ntargets:\n- kind: layer\n  name: foo\n  dependencies:\n  - kind: pip\n  - kind: pipper\n  - kind: poetry\n```\n\nCurrently `pip`, `pipper` and `poetry` package managers are supported. For any of the\npackage managers, the dependencies can be specified explicitly with the \n`package(s)` key.\n\n```yaml\ntargets:\n- kind: layer\n  name: foo\n  dependencies:\n  - kind: pip\n    packages: \n    - spam\n    - hamd\n  - kind: pipper\n    package: spammer\n```\n\nIt\'s also possible to specify a file to where the package dependencies\nhave been defined.\n\n```yaml\ntargets:\n- kind: layer\n  name: foo\n  dependencies:\n  - kind: pip\n    file: requirements.layer.txt\n  - kind: pipper\n    file: pipper.layer.json\n```\n\nIf no packages or file is specified, the default file for the given package\nmanager will be used by default (e.g. `requirements.txt` for pip,\n `pipper.json` for pipper, and `pyproject.toml` for poetry).\n\nIt is also possible to specify the same kind of package manager multiple\ntimes in this list to aggregate dependencies from multiple locations.\n\n### targets[N].dependencies(kind="pipper")\n\nPipper repositories have additional configuration not associated with pip\npackages. To support pipper libraries, there are two additional attributes\nthat can be specified: `bucket` and `prefix`.\n\nThe `bucket` is required as it specifies the S3 bucket used as the package\nsource and should be read-accessible by the profile invoking reviser.\n\nThe `prefix` is an optional alternate package prefix within the S3 bucket.\nUse this only if you are using an alternate prefix with for your pipper\npackage.\n\n```yaml\ntargets:\n- kind: layer\n  name: foo\n  dependencies:\n  - kind: pipper\n    file: pipper.layer.json\n    bucket: bucket-name-where-pipper-package-resides\n    prefix: a/prefix/that/is/not/just/pipper\n```\n\n### targets[N].dependencies(kind="poetry")\n\nPoetry repositories have additional `extras` configuration that can be used to\nspecify optional dependency groups to install in the lambda. This can be useful\nto separate dependencies by function.\n\n```yaml\ntargets:\n- kind: layer\n  name: foo\n  dependencies:\n  - kind: poetry\n    extras:\n    - group\n```\n\n### targets[N].bundle\n\nThe target bundle object contains the attributes that define the bundle\nthat will be created and uploaded to the functions or layers in a given\ntarget as part of the deployment process. It\'s primary purpose is to define\nwhat files should be included in the bundling process, which it achieves\nwith the following attributes.\n\n#### targets[N].bundle.include(s)\n\nThe `include(s)` key is a string or list of Python glob-styled includes\nto add to the bundle. If no includes are specified, the default behavior is:\n\n- **function targets**: copy the first directory found that contains an \n  *__init__.py* file.\n- **layer targets**: do not copy anything and assume dependencies are the\n  only files to copy into the bundle.\n\nAll paths should be referenced relative to the root path where the\n`lambda.yaml` is located. For a recursive matching pattern, the glob syntax\nshould be used as `**/*.txt` or if restricted to a folder inside of the root\ndirectory then `folder/**/*.txt`. To include the entire contents of a\ndirectory, specify the path to the folder.\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  bundle:\n    includes:\n    # This is shorthand for "foo_library/**/*"\n    - foo_library\n    # All Python files in the "bin/" folder recursively.\n    - bin/**/*.py\n    # All Jinja2 files in the root directory that begin "template_".\n    - template_*.jinja2\n```\n\n#### targets[N].bundle.exclude(s)\n\nThe `exclude(s)` key is an optional one that is also a string or list of\nPython glob-styled paths to remove from the matching `include(s)`. These\nare applied to the files found via the includes and do not need to be\ncomprehensive of all files in the root directory. Building on the example\nfrom above:\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  bundle:\n    includes:\n    # This is shorthand for "foo_library/**/*"\n    - foo_library\n    # All Python files in the "bin/" folder recursively.\n    - bin/**/*.py\n    # All Jinja2 files in the root directory that begin "template_".\n    - template_*.jinja2\n    exclues:\n    - template_local.jinja2\n    - template_testing.jinja2\n```\n\nThis example would remove two of the template file matches from the includes\nfrom the files copied into the bundle for deployment.\n\nAll `__pycache__`, `*.pyc` and `.DS_Store` files/directories are\nexcluded from the copying process in all cases and do not need to be \nspecified explicitly.\n\n#### targets[N].bundle.omit_package(s)\n\nThere can be cases where dependencies install dependencies of their own that\nyou may not want copied over to the bundle. The most common case is a\ndependency that requires `boto3`, which is available by default in lambda\nfunctions already. In that case it can be useful to list site packages that\nshould not be copied into the bundle but may have been installed as a side\neffect of the dependency installation process.\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  bundle:\n    omit_package: boto3\n  dependencies:\n  - kind: pip\n    # Installs a package that requires boto3, which is therefore installed\n    # into the site-packages bundle directory as a result.\n    # https://github.com/awslabs/aws-lambda-powertools-python\n    package: aws-lambda-powertools\n```\n\nIn the above example `aws-lambda-powertools` causes `boto3` to be installed\nas well. However, since lambda functions have `boto3` installed by default,\nit\'s possible to omit that package from the bundling process so that it isn\'t\ninstalled twice.\n\nNote, however, that installing `boto3` directly in a bundle can be beneficial\nbecause it gives you the ability to install the version that is compatible\nwith your given source code and dependencies. The `boto3` version on the lambda\nfunction can be aged and stale.\n\n#### targets[N].bundle.handler\n\nThis attribute only applies to function targets and gives the location of\nfile and function entrypoint for the lambda function(s) in the target. The\nformat matches the expected value for lambda functions, which is \n`<filename_without_extension>.<function_name>`.\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  bundle:\n    handler: function:main\n```\n\nIn this case the bundler would expect to find `function.py` in the top-leve\ndirectory alongside `lambda.yaml` and inside it there would be a \n`main(event, context)` function that would be called when the function(s)\nare invoked.\n\nIf this value is omitted, the default value of `lambda_function.lambda_handler`\nwill be used as this matches the AWS lambda Python function documentation.\n\n## function targets\n\nIn addition to the common attributes described above that are shared between\nboth function and layer targets, there are a number of additional\nattributes that apply only to function targets. These are:\n\n### (function) targets[N].layer(s)\n\nSpecifies one or more layers that should be attached to the targeted\nfunction(s). Layers can be specified as fully-qualified ARNs for externally\nspecified layers, e.g. a layer created in another AWS account, or by name\nfor layers specified within the account and layers defined within the targets\nof the configuration file.\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  layer: arn:aws:lambda:us-west-2:999999999:layer:bar\n  ...\n```\n\nor for multiple layers:\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  layers:\n  # A layer defined in another account is specified by ARN.\n  - arn:aws:lambda:us-west-2:999999999:layer:bar\n  # A layer in this account is specified by name. This layer may also be\n  # a target in this configuration file.\n  - baz\n  ...\n- kind: layer\n  name: baz\n  ...\n```\n\nBy default, deployments will use the latest available version of each layer,\nbut this can be overridden by specifying the layer ARN with its version:\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  layer: arn:aws:lambda:us-west-2:999999999:layer:bar:42\n  ...\n```\n\nIn the above example the layer will remain at version 42 until explicitly\nmodified in the configuration file.\n\nLayers can also be defined as objects instead of attributes. The two-layer\nexample from above could be rewritten as:\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  layers: \n  - arn: arn:aws:lambda:us-west-2:999999999:layer:bar\n  - name: baz\n  ...\n```\n\nWhen specified as an object with attributes, there are a number of additional\nattributes that can be specified as well. First, `version` can be specified\nas a separate key from the arn or name, which in many cases can make it easier\nto work with than appending it to the end of the arn or function itself for\nprogrammatic/automation:\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  layers: \n  - arn: arn:aws:lambda:us-west-2:999999999:layer:bar\n    version: 42\n  - name: baz\n    version: 123\n  ...\n```\n\nNext is that the layer objects accept `only` and `except` keys that can be\nused to attach the layers to certain functions in the target and not others.\nThis can be useful in cases where development and production targets share\na lot in common, but perhaps point to different versions of a layer or perhaps\nseparate development and production layers entirely. It can also be useful\nwhen a target of functions share a common codebase but don\'t all need the\nsame dependencies. For performance optimization, restricting the layer \ninclusions only to those that need the additional dependencies can be\nbeneficial.\n\nThe `only` and `except` attributes can be specified as a single string\nor a list of strings that match against *unix pattern matching*. For example,\nexpanding on the example from above:\n\n```yaml\ntargets:\n- kind: function\n  names: \n  - foo-devel\n  - foo-devel-worker\n  - foo-prod\n  - foo-prod-worker\n  layers: \n  - name: baz-devel\n    only: foo-devel*\n  - name: baz-devel-worker\n    only: foo-devel-worker\n  - name: baz-prod\n    only: foo-prod*\n  - name: baz-prod-worker\n    only: foo-prod-worker\n  ...\n```\n\nthis example shows 4 layers that are conditionally applied using the only\nkeyword. The example could be rewritten with the `except` key instead:\n\n```yaml\ntargets:\n- kind: function\n  names: \n  - foo-devel\n  - foo-devel-worker\n  - foo-prod\n  - foo-prod-worker\n  layers: \n  - name: baz-devel\n    except: foo-prod*\n  - name: baz-devel-worker\n    except:\n    - foo-prod*\n    - foo-devel\n  - name: baz-prod\n    except: foo-devel*\n  - name: baz-prod-worker\n    except:\n    - foo-devel*\n    - foo-prod\n  ...\n```\n\nAnd either way works. The two (`only` and `except`) can also be combined\nwhen that makes more sense. For example, the `baz-devel-worker` from above\ncould also be written as:\n\n```yaml\n  - name: baz-devel-worker\n    only: foo-devel*\n    except: foo-devel\n```\n\nNote that if `only` is specified it is processed first and then `except` is\nremoved from the matches found by `only`. \n\n### (function) targets[N].memory\n\nThis specifies the function memory in megabytes either as an integer or\na string with an `MB` suffix.\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  memory: 256MB \n```\n\n### (function) targets[N].timeout\n\nThis specifies the function timeout in seconds either as an integer or\na string with an `s` suffix.\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  timeout: 42s \n```\n\n### (function) targets[N].variable(s)\n\nVariables contains a list of environment variables to assign to the function.\nThey can be specified simply with as a string `<KEY>=<value>` syntax:\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  variable: MODE=read-only\n```\n\nHere a single environment variable is specified that maps `"MODE"` to the\nvalue *"ready-only"*. A more programmatic-friendly way is to specify the\nname and value as attributes of a variable:\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  variables: \n  - name: MODE\n    value: read-only\n```\n\nSome environment variables may be managed through other means, e.g. \nterraform that created the function in the first place or another command\ninterface used to update the function. For those cases, the `preserve`\nattribute should be set to true and no value specified.\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  variables: \n  - name: MODE\n    preserve: true\n```\n\nIn this case the `MODE` environment variable value will be preserved between\nfunction deployments to contain the value that was already set.\n\nFinally, variables support the same `only` and `exclude` attributes that\nare found for target layers so that environment variables can be specified\ndifferently for subsets of targets.\n\nThe `only` and `except` attributes can be specified as a single string\nor a list of strings that match against *unix pattern matching*. For example,\nexpanding on the example from above:\n\n```yaml\ntargets:\n- kind: function\n  names: \n  - foo-prod\n  - foo-devel\n  variables: \n  - name: MODE\n    value: write\n    only: \'*prod\'\n  - name: MODE\n    value: read-only\n    except: \'*prod\'\n```\n\n### (function) targets[N].ignore(s)\n\nIgnores allows you to specify one or more configuration keys within a function\ntarget that should be ignored during deployments. For cases where any of the \nconfiguration values:\n\n- `memory`\n- `timeout`\n- `variables`\n\nare managed by external systems, they can be specified by the ignores to\nprevent changes being applied by reviser.\n\n```yaml\ntargets:\n- kind: function\n  name: foo\n  ignores:\n  - memory\n  - timeout\n```\n\n## run\n\nThe run attribute contains an optional object of batch non-interactive commands\nto run when the shell is called with that run key. This is useful for \norchestrating actions for CI/CD purposes as the commands will be processed\nwithin a shell environment without user prompts and then the shell will exit\nwhen complete without waiting for additional input.\n\n```yaml\nrun:\n  deploy-prod:\n  - select function *prod \n  - push --description="($CI_COMMIT_SHORT_SHA): $CI_COMMIT_TITLE"\n  - alias test -1\ntargets:\n- kind: function\n  names:\n  - foo-prod\n  - foo-devel\n```\n\nIn the example above, the `deploy-prod` run command macro/group would start\nthe shell and then non-interactively execute the three commands in order\nto first select the *foo-prod* function, then to build and deploy that function\nwith a description created from CI environment variables and finally move the\n*test* alias to the newly deployed version using a negative version index of\n*-1*. After those three commands are executed reviser will exit the\nshell automatically, successfully ending that process.\n\nThere is also a special `shell` command that can be used in run command\nmacros/groups that will start the shell in interactive mode. This is useful\nfor using run command macros/groups for pre-configuration during startup of\nthe interactive shell. Building on the previous example,\n\n```yaml\nrun:\n  deploy-prod:\n  - select function *prod \n  - push --description="($CI_COMMIT_SHORT_SHA): $CI_COMMIT_TITLE"\n  - alias test -1\n  devel:\n  - select * *devel\n  - bundle\n  - shell\ntargets:\n- kind: function\n  names:\n  - foo-prod\n  - foo-devel\n- kind: layer\n  names:\n  - bar-devel\n  - bar-prod\n```\n\nhere we\'ve added a `devel` run command macro/group that will select the devel\nfunction and layer and bundle those but not deploy them. After that\'s complete\nthe shell command will kick off the interactive session and ask for user\ninput. The benefit of this particular run command macro/group is to select\nthe development targets and pre-build them to cache the dependencies for the\nshell user while they continue to develop and deploy the source code to the\nfunction.\n',
    'author': 'Scott Ernst',
    'author_email': 'swernst@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/rocket-boosters/reviser',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import typing


def to_human_readable_size(
    number_of_bytes: typing.Optional[int],
) -> typing.Optional[str]:
    """Converts bytes value to human readable size."""
    if number_of_bytes is None:
        return None

    step_to_greater_unit = 1024
    unit = "bytes"
    value = float(number_of_bytes)

    if value < 1024:
        return f"{int(value)} {unit}"

    if (value / step_to_greater_unit) >= 1:
        value /= step_to_greater_unit
        unit = "KB"

    if (value / step_to_greater_unit) >= 1:
        value /= step_to_greater_unit
        unit = "MB"

    if (value / step_to_greater_unit) >= 1:
        value /= step_to_greater_unit
        unit = "GB"

    if (value / step_to_greater_unit) >= 1:
        value /= step_to_greater_unit
        unit = "TB"

    value = round(value, 1)
    return f"{value} {unit}"



import dataclasses
import fnmatch
import pathlib
import re
import tempfile
import typing

from botocore.client import BaseClient

from reviser.definitions import abstracts
from reviser.definitions import configurations
from reviser.definitions import enumerations
from reviser.definitions import selections


def _is_selected_match(
    target: "Target",
    name: str,
):
    """
    Determines whether or not the target name is a match for the
    current selection set on the target object.
    """
    if not target.selection:
        return True

    if exacts := getattr(target.selection, f"{target.kind.value}_names"):
        return name in exacts

    needles = getattr(target.selection, f"{target.kind.value}_needles")
    any_needles = bool(target.selection.function_needles) or bool(
        target.selection.layer_needles
    )

    return (
        # Match everything if `--all` was specified.
        target.selection.bundle_all
        # No needles specified.
        or not any_needles
        # Identical match.
        or name in needles
        # Shell-style wildcard matches.
        or any([fnmatch.fnmatch(name, n) for n in needles])
        # Partial character matches.
        or any([n.lower() in name.lower() for n in needles])
    )


@dataclasses.dataclass(frozen=True)
class Target(abstracts.Specification):
    """
    Function or layer definition data structure where a target can
    represent one or more functions or one or more layers in the same
    definition for cases where function/layer duplication is desirable.
    A target, however, is of a homogeneous kind, meaning it will only
    be a function or layer.
    """

    configuration: "configurations.Configuration"
    selection: typing.Optional["selections.Selection"] = None

    @property
    def aws_region(self) -> str:
        """AWS region name where this target resides."""
        return self.get("region", default=self.configuration.aws_region)

    @property
    def bucket(self) -> typing.Optional[str]:
        """Retrieves the bucket to use for uploading to S3."""
        buckets = self.get_first(
            ["buckets"],
            ["bucket"],
            default=self.configuration.bucket,
        )

        if not buckets:
            return None

        if isinstance(buckets, str):
            return buckets

        return buckets[self.connection.aws_account_id]

    @property
    def bundle(self) -> "configurations.Bundle":
        """Bundle object associated with this target configuration."""
        return configurations.Bundle(
            directory=self.directory,
            data=self.get("bundle", default={}),
            connection=self.connection,
            target=self,
        )

    @property
    def layer_attachments(self) -> typing.List["configurations.AttachedLayer"]:
        """
        List of layers that should be attached to functions in this target.
        This will always be an empty list for layer targets.
        """
        if self.kind == enumerations.TargetType.LAYER:
            return []
        values = self.get_first_as_list(["layers"], ["layer"], default=[])
        values = [{"name": v} if isinstance(v, str) else v for v in values]
        return [
            configurations.AttachedLayer(
                directory=self.directory,
                data=v,
                connection=self.connection,
                target=self,
            )
            for v in values
        ]

    @property
    def variables(self) -> typing.List["configurations.EnvironmentVariable"]:
        """
        List of environment variables that will be applied during
        function configuration updates within the deploy action. This
        will always be an empty list for layer targets that don't support
        environment variables.
        """
        if self.kind == enumerations.TargetType.LAYER:
            return []
        values = self.get_first_as_list(["variables"], ["variable"], default=[])
        values = [{"arg": v} if isinstance(v, str) else v for v in values]
        return [
            configurations.EnvironmentVariable(
                directory=self.directory,
                data=v,
                connection=self.connection,
                target=self,
            )
            for v in values
        ]

    @property
    def ignores(self) -> typing.List[str]:
        """
        Items that should be skipped when modifying function configurations
        during the deployment process.
        """
        ignores = self.get_first_as_list(["ignores"], ["ignore"], default=[])
        return [i.lower() for i in ignores]

    @property
    def bundle_directory(self) -> pathlib.Path:
        """Temporary location where the target will be assembled."""
        return pathlib.Path(tempfile.gettempdir()).joinpath(self.uuid)

    @property
    def bundle_zip_path(self) -> pathlib.Path:
        """Temporary location where the target zip bundle will be saved."""
        return pathlib.Path(tempfile.gettempdir()).joinpath(f"{self.uuid}.zip")

    @property
    def site_packages_directory(self) -> pathlib.Path:
        """Temporary location where site packages will be installed."""
        if self.kind == enumerations.TargetType.LAYER:
            return self.bundle_directory.joinpath("python")
        return self.bundle_directory.joinpath("site_packages")

    @property
    def kind(self) -> "enumerations.TargetType":
        """Target type as one of 'function' or 'layer'."""
        # noinspection PyArgumentList
        return enumerations.TargetType(
            value=self.get(
                "kind",
                default=enumerations.TargetType.FUNCTION.value,
            )
        )

    @property
    def names(self) -> typing.List[str]:
        """Names of the targets associated with this definition."""
        names = self.get_first_as_list(["names"], ["name"], default=[])
        return [str(n) for n in names if n and _is_selected_match(self, n)]

    @property
    def timeout(self) -> typing.Optional[int]:
        """
        Timeout, in seconds, after which the function stops executing. This
        will always be None for layers.
        """
        value = self.get("timeout")
        if not value or self.kind == enumerations.TargetType.LAYER:
            return None

        try:
            return int(value)
        except ValueError:
            regex = re.compile(r"^(?P<value>\d+).*")
            if (match := regex.match(value)) is not None:
                return int(match.groupdict()["value"])
            return None

    @property
    def memory(self) -> typing.Optional[int]:
        """
        Memory, in MB, available to the function when executing. This
        will always be None for layers.
        """
        value = self.get("memory")
        if not value or self.kind == enumerations.TargetType.LAYER:
            return None

        try:
            return int(value)
        except ValueError:
            regex = re.compile(r"^(?P<value>\d+).*")
            if (match := regex.match(value)) is not None:
                return int(match.groupdict()["value"])
            return None

    @property
    def dependencies(self) -> typing.Tuple["configurations.Dependency", ...]:
        """Dependencies for this function definition."""
        output = []

        dependency: configurations.Dependency
        dependency_types = enumerations.DependencyType
        mappings: typing.Dict[str, typing.Callable] = {
            dependency_types.PIP.value: configurations.PipDependency,
            dependency_types.PIPPER.value: configurations.PipperDependency,
            dependency_types.POETRY.value: configurations.PoetryDependency,
        }
        for data in self.get("dependencies", default=[]):
            kind = data.get("kind", dependency_types.PIP.value)
            constructor = mappings.get(kind, configurations.PipDependency)
            output.append(
                constructor(
                    directory=self.directory,
                    data=data,
                    connection=self.connection,
                    target=self,
                )
            )

        return tuple(output)

    def ignores_any(self, *args: str) -> bool:
        """True if any of the specified args appear in the ignores list."""
        ignores = self.ignores
        finder = (True for a in args if a in ignores)
        return next(finder, False)

    def client(self, service_name: str) -> BaseClient:
        """
        Creates a boto3 client for the given service configured for use
        when acting on this target.
        """
        return self.connection.client(service_name, region_name=self.aws_region)

    def serialize(self) -> dict:
        """Serializes the object for output representation."""
        if self.kind == enumerations.TargetType.FUNCTION:
            values = {
                "layers": [a.serialize() for a in self.layer_attachments],
                "variables": [v.serialize() for v in self.variables],
                "memory": self.memory,
                "timeout": self.timeout,
                "ignores": self.ignores,
            }
        else:
            values = {}

        return {
            "kind": self.kind.value,
            "names": self.names,
            "region": self.aws_region,
            "bundle_directory": str(self.bundle_directory),
            "bundle_zip_path": str(self.bundle_zip_path),
            "site_packages_directory": str(self.site_packages_directory),
            "bundle": self.bundle.serialize(),
            "dependencies": [d.serialize() for d in self.dependencies],
            **values,
        }



import copy
import os
import pathlib
import shutil
import subprocess
import typing

from reviser import definitions


def _install_pip_package(name: str, site_packages: pathlib.Path):
    """Installs the specified pip package."""
    cmd = [
        "python",
        "-m",
        "pip",
        "install",
        "--upgrade",
        name,
        "-t",
        f'"{site_packages}"',
    ]
    print(" ".join(cmd).replace(" -", "\n  -"))
    os.system(" ".join(cmd))


def _install_pipper_package(
    name: str,
    site_packages: pathlib.Path,
    env: dict,
    arguments: typing.List[str] = None,
):
    """Installs the specified pipper package."""
    cmd = [
        "pipper",
        "install",
        name,
        "--upgrade",
        "--target",
        f'"{site_packages}"',
        *(arguments or []),
    ]
    print(" ".join(cmd).replace(" -", "\n  -"))
    subprocess.run(cmd, env=env, check=True)


def _install_poetry(dependency: "definitions.PoetryDependency"):
    """Installs poetry dependencies in the target's site packages."""
    for package in dependency.get_package_names() or []:
        print(f'\n[INSTALLING]: "{package}" poetry package')
        _install_pip_package(
            package,
            dependency.target.site_packages_directory,
        )
        print(f'\n[INSTALLED]: "{package}" poetry package')


def _install_pip(dependency: "definitions.PipDependency"):
    """Installs poetry dependencies in the target's site packages."""
    for package in dependency.get_package_names() or []:
        print(f'\n[INSTALLING]: "{package}" pip package')
        _install_pip_package(
            package,
            dependency.target.site_packages_directory,
        )
        print(f'\n[INSTALLED]: "{package}" pip package')


def _install_pipper(dependency: "definitions.PipperDependency"):
    """
    Installs any pipper dependencies alongside the standard pip site packages
    at the specified bundle site packages directory location.
    """
    credentials = dependency.connection.get_credentials()
    env_vars = {
        "PIPPER_AWS_ACCESS_KEY_ID": credentials.access_key,
        "PIPPER_AWS_SECRET_ACCESS_KEY": credentials.secret_key,
        "PIPPER_AWS_SESSION_TOKEN": credentials.token,
    }
    env = copy.copy(os.environ)
    # Depending on what type os session source is used, one or more of
    # these will be None and None values cannot be included in environment
    # variables without raising a TypeError: str expected, not NoneType.
    env.update({k: v for k, v in env_vars.items() if v is not None})

    for package in dependency.get_package_names():
        additional_kwargs = {
            "--bucket": dependency.bucket,
            "--prefix": dependency.prefix,
        }
        arguments = [f"{k}={v}" for k, v in additional_kwargs.items() if v is not None]

        print(f'\n[INSTALLING]: "{package}" pipper package')

        _install_pipper_package(
            package,
            dependency.target.site_packages_directory,
            typing.cast(dict, env),
            arguments,
        )
        print(f'\n[INSTALLED]: "{package}" pipper package')


def install_dependencies(target: "definitions.Target"):
    """Install the dependencies for the specified target."""
    if target.site_packages_directory.exists():
        print("\n[RESET]: Site packages directory exists and is being reset.")
        shutil.rmtree(
            str(target.site_packages_directory),
            ignore_errors=True,
        )

    target.site_packages_directory.mkdir(exist_ok=True, parents=True)

    callers: typing.Any = {
        definitions.DependencyType.PIPPER: _install_pipper,
        definitions.DependencyType.PIP: _install_pip,
        definitions.DependencyType.POETRY: _install_poetry,
    }

    for dependency in target.dependencies:
        # noinspection PyTypeChecker
        callers[dependency.kind](dependency)



"""
Uploads the bundled contents to the upload S3 bucket and then publishes
a new version of each of the lambda targets with that new bundle and
any modified settings between the current configuration and that target's
existing configuration. This command will fail if a target being deployed
has not already been bundled.
"""
import argparse
import os
import string
import typing

from reviser import deploying
from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Shell auto-completes for this command."""
    return ["--description", "--dry-run"]


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    parser.add_argument(
        "--description",
        help="""
        Specify a message to assign to the version published by
        the deploy command.
        """,
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="""
        If set, the deploy operation will be exercised without
        actually carrying out the actions. This can be useful to
        validate the deploy process without side effects.
        """,
    )


def run(ex: "interactivity.Execution"):
    """Execute a bundle operation on the selected function/layer targets."""
    description = string.Template(ex.args.get("description") or "").substitute(
        os.environ
    )
    print("\n\n")
    deployed_targets = deploying.deploy(
        context=ex.shell.context,
        selection=ex.shell.selection,
        description=description,
        dry_run=ex.args.get("dry_run") or False,
    )
    print("\n")
    return ex.finalize(
        status="DEPLOYED",
        message="Selected items have been deployed.",
        info={
            "dry_run": ex.args.get("dry_run"),
            "items": [n for t in deployed_targets for n in t.names],
            "description": description,
        },
        echo=True,
    )



"""
Combined single command for bundling and deploying the selected targets.
"""
import argparse
import typing

from reviser import interactivity
from ..commands import bundler
from ..commands import deployer


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Shell auto-completes for this command."""
    return bundler.get_completions(completer) + deployer.get_completions(completer)


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    bundler.populate_subparser(parser)
    deployer.populate_subparser(parser)


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Execute a bundle operation on the selected function/layer targets."""
    out = bundler.run(ex)
    if out.result.status != "BUNDLED":
        return out
    return deployer.run(ex)
